// Variable Class
// This class handles memory management for variables and pointers. There are two
// types of variables: values and references. These correspond to r-values and
// l-values respectively in C. The v() function will access the variable's
// r-value, automatically de-referencing if necessary. An error will occur if an
// attempt is made to access the value of an uninitialized reference. Each variable
// keeps track of how many references are being made to it. An error will occur if 
// the destructor is called on a variable that is still in use.
// These classes are designed to prevent illegal pointer operations.

#ifndef __VAR_H
#define __VAR_H

#include <stdlib.h>
#include <stdio.h>

template<class T>
class var {
public:
	int uses;	// Number of references being made to this variable

	var() { uses = 0;}
	~var() {
		if(uses > 0) {
			printf("Attempt to delete variable in use");
			exit(1);
		}
	}
	virtual T& v() = 0;				// Returns modifiable value
	virtual const T& c() const = 0;	// Returns const value
	operator T&() { return v();}	// Cast operators
	operator T*() { return &v();}
	int InUse() const { return uses; }
};

template<class T>
class val : public var<T> {
	T value;
public:
	val(const val<T>& init) : value(init.c()) {}	// Copy constructors
	val(const var<T>& init) : value(init.c()) {}	
	val(const T& init) : var<T>(), value(init) {}
	T& operator=(const T& init) { 
		value = init;
		return value;
	}
	T& v() { return value;}
	const T& c() const { return value;}
};

template<class T>
class ref : public var<T> {
	var<T> *ptr;
public:
	ref() : var<T>() { ptr = (var<T>*)NULL;}
	~ref() { PointTo();}
	ref(ref<T>& p) {			// Copy constructors
		ptr = (var<T>*)NULL;
		PointTo(p);
	}
	ref(var<T>& p) {
		ptr = (var<T>*)NULL;
		PointTo(p);
	}
	T& operator=(const T& init) { 
		v() = init; 
		return v();
	}
	operator T&() { return v();}
	T& v() {     // Retrieves the value by dereferencing
		if(ptr) return ptr->v();
		printf("Attempt to access uninitialized pointer");
		exit(2);
		// This code never reached - prevents compiler error about the return value
		return (*(T*)NULL);
	}
	const T& c() const {	// Retrieves const value by dereferencing
		if(ptr) return ptr->c();
		printf("Attempt to access uninitialized pointer");
		exit(2);
		// This code never reached - prevents compiler error about the return value
		return (*(T*)NULL);
	}
	void PointTo(var<T>& p) {   // Sets reference to point to a variable
		PointTo();
		ptr = &p;
		ptr->uses++;
	}
	void PointTo() {      // Causes reference to point to nothing
		if(IsNull()) return;
		ptr->uses--;
		ptr = (var<T>*)NULL;
	}
	int IsNull() const { return (ptr == (var<T>*)NULL);}
};

#endif
